import loopy as lp
import numpy as np
import pymbolic.primitives as prim
from dune.perftool.loopy.temporary import DuneTemporaryVariable
from dune.perftool.loopy.symbolic import substitute
from dune.perftool.tools import get_pymbolic_indices


def add_vcl_vector(knl, iname):
    read_insn_to_vec_instance = dict()
    write_insn_to_vec_instance = dict()
    read_vecs = set()
    write_vecs = set()

    from loopy.symbolic import WalkMapper
    class FindAlias(WalkMapper):
        def __init__(self, id, expr, vecs, insn_to_vec):
            self.id = id
            self.vecs = vecs
            self.insn_to_vec = insn_to_vec
            self.rec(expr)

        def post_visit(self,expr):
            if isinstance(expr, prim.Subscript) \
                    and isinstance(expr.aggregate, prim.Variable) \
                    and expr.aggregate.name.endswith('alias'):
                vec_name = expr.aggregate.name.replace('alias', 'vec')
                self.vecs.add(vec_name)
                self.insn_to_vec.setdefault(self.id, set())
                self.insn_to_vec[self.id].add((vec_name,expr))
            return
    # find all instances where an alias is read from or written to
    # save the name of the corresponding vector and the alias subscript expression
    for insn in knl.instructions:
        if isinstance(insn, lp.MultiAssignmentBase):
            FindAlias(insn.id, insn.expression, read_vecs, read_insn_to_vec_instance)
            if isinstance(insn, lp.Assignment):
                FindAlias(insn.id, insn.assignee, write_vecs, write_insn_to_vec_instance)

    read_insns = frozenset(read_insn_to_vec_instance.keys())
    write_insns = frozenset(write_insn_to_vec_instance.keys())

    # add new temporaries for vectors
    new_vec_temporaries = dict()
    for vec in read_vecs|write_vecs:
        new_vec_temporaries[vec] = DuneTemporaryVariable(vec, dtype=np.float64, shape=(4,), managed=True,
                                                         scope=lp.temp_var_scope.PRIVATE, dim_tags=('vec',))

    # add load instructions if the vector is read, based on the read instruction
    # TODO brauche mehrere vectoren, falls in einer insn von einem alias mit unterschiedlichen idx gelesen wird
    # muss dafür eigentlich nur den namen des vectors anpassen
    load_insns = []
    read_dependencies = dict()
    for id in read_insns:
        # names
        for vec, expr in read_insn_to_vec_instance[id]:
            alias = vec.replace('vec', 'alias')

            # flat index without vec iname
            strides = tuple(tag.stride for tag in knl.temporary_variables[alias].dim_tags)
            index = prim.Sum(tuple(prim.Product(z) for z in zip(substitute(expr, {iname:0}).index_tuple, strides)))

            # add load instruction
            code = "{}.load({} + {});".format(vec, alias, index)
            load_id = knl.make_unique_instruction_id(insns=knl.instructions+load_insns, based_on='insn_'+vec+'_load')
            load_insns.append(lp.CInstruction(iname_exprs=[], code=code, read_variables=frozenset({alias}),
                                              within_inames=knl.id_to_insn[id].within_inames|frozenset(get_pymbolic_indices(expr)),
                                              #assignees=(lp.TaggedVariable(vec, 'coeff_vec'), ), #sonst denkt looy das müsste ein array von Vec4d sein...
                                              id=load_id))
            read_dependencies.setdefault(id, set())
            read_dependencies[id].add(load_id)

    # add store instructions if the vector is written, based on the write instruction
    store_insns = []
    for id in write_insns:
        for vec, expr in write_insn_to_vec_instance[id]:
            alias = vec.replace('vec', 'alias')

            # flat index without vec iname
            strides = tuple(tag.stride for tag in knl.temporary_variables[alias].dim_tags)
            from pudb import set_trace; set_trace()
            index = prim.Sum(tuple(prim.Product(z) for z in zip(substitute(expr, {iname:0}).index_tuple, strides)))

            # add store instruction
            code = "{}.store({} + {});".format(vec, alias, index)
            store_id = knl.make_unique_instruction_id(insns=knl.instructions+load_insns, based_on='insn_'+vec+'_store')
            store_insns.append(lp.CInstruction(iname_exprs=[], code=code, read_variables=frozenset({alias}),
                                               within_inames=knl.id_to_insn[id].within_inames|frozenset(get_pymbolic_indices(expr)),
                                               depends_on=knl.id_to_insn[id].depends_on
                                                          |frozenset({id})
                                                          |read_dependencies[id],
                                               id=store_id))

    # exchange alias for vector
    new_insns = []
    for insn in knl.instructions:
        if not insn.id in read_insns|write_insns:
            new_insns.append(insn)
        else:
            subst_map = dict()
            for vec, expr in read_insn_to_vec_instance[insn.id]:
                subst_map[expr] = prim.Subscript(prim.Variable(vec), (prim.Variable(iname),))
            new_insn = insn
            if insn.id in read_insns:
                new_insn = new_insn.copy(expression=substitute(insn.expression, subst_map),
                                         depends_on=insn.depends_on|read_dependencies[insn.id])
            if insn.id in write_insns:
                new_insn = new_insn.copy(assignee=substitute(insn.assignee, subst_map))
            new_insns.append(new_insn)

    from loopy.kernel.creation import resolve_dependencies
    return resolve_dependencies(knl.copy(instructions=new_insns+load_insns+store_insns,
                                         temporary_variables=dict(**knl.temporary_variables, **new_vec_temporaries)))


def vectorize_micro_elements(knl):
    if "subel_x" in knl.all_inames():
        knl = lp.split_iname(knl,"subel_x",4, inner_tag='vec')
        knl = add_vcl_vector(knl,'subel_x_inner')
    return knl
